!MNH_LIC Copyright 1994-2014 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt  
!MNH_LIC for details. version 1.
!     ###########################
      MODULE MODI_ADVECUVW_WENO_K
!     ###########################
!
INTERFACE
!
      SUBROUTINE ADVECUVW_WENO_K(HLBCX, HLBCY, KWENO_ORDER, PUT, PVT, PWT,       &
                             PRUCT, PRVCT, PRWCT, PRUS, PRVS, PRWS, TPHALO2LIST  &
#ifndef _OPENACC
                             )
#else
                             , ZMEAN, ZWORK)
#endif
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
INTEGER,                  INTENT(IN)    :: KWENO_ORDER   ! Order of the WENO
                                                         ! scheme (3 or 5)
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRUCT ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVCT !  components
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRWCT ! of momentum
!$acc declare present(PRUCT,PRVCT,PRWCT)
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PUT, PVT, PWT        ! U,V,W at t
!$acc declare present(PUT,PVT,PWT)
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS, PRVS, PRWS     ! Source terms
!$acc declare present(PRUS,PRVS,PRWS)
!
TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list for diffusion
!
#ifdef _OPENACC
! Work arrays
REAL, DIMENSION(SIZE(PUT,1), SIZE(PUT,2), SIZE(PUT,3)) :: ZMEAN, ZWORK
!$acc declare present(ZMEAN,ZWORK)
#endif
!
END SUBROUTINE ADVECUVW_WENO_K
!
END INTERFACE
!
END MODULE MODI_ADVECUVW_WENO_K
!
!     ##########################################################################
      SUBROUTINE ADVECUVW_WENO_K(HLBCX, HLBCY, KWENO_ORDER, PUT, PVT, PWT,      &
                             PRUCT, PRVCT, PRWCT, PRUS, PRVS, PRWS, TPHALO2LIST &
#ifndef _OPENACC
                             )
#else
                             , ZMEAN, ZWORK)
#endif
!     ##########################################################################
!
!!    AUTHOR
!!    ------
!!
!!
!!    MODIFICATIONS
!!    ------------- 
!!      J.Escobar 21/03/2013: for HALOK comment all NHALO=1 tests
!!		  T.Lunet	 	02/10/2014: add get_halo for WENO 5
!!					suppress comment of NHALO=1 tests
!!      J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1 
!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODE_ll
!
USE MODD_PARAMETERS
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
#ifndef _OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_ADVEC_WENO_K_1_AUX
USE MODI_ADVEC_WENO_K_2_AUX
USE MODI_ADVEC_WENO_K_3_AUX
!
USE MODD_CONF,        ONLY : NHALO
USE MODE_MPPDB
USE MODI_GET_HALO
!
#ifdef _OPENACC
USE MODE_DEVICE
USE MODE_MNH_ZWORK, ONLY : ZT3D, MNH_GET_ZT3D , MNH_REL_ZT3D
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
INTEGER,                  INTENT(IN)    :: KWENO_ORDER   ! Order of the WENO
                                                         ! scheme (3 or 5)
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRUCT  ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRVCT  !  components
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRWCT  ! of momentum
!$acc declare present(PRUCT,PRVCT,PRWCT)
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PUT, PVT, PWT     ! Variables at t
!$acc declare present(PUT,PVT,PWT)
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS, PRVS, PRWS     ! Source terms
!$acc declare present(PRUS,PRVS,PRWS)
!
TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list for diffusion
!
!*       0.2   Declarations of local variables :
!
TYPE(HALO2LIST_ll), POINTER :: TZHALO2_UT,TZHALO2_VT,TZHALO2_WT

TYPE(LIST_ll), POINTER :: TZHALO2_ZMEAN
INTEGER                     :: IINFO_ll    ! return code of parallel routine
!
REAL, DIMENSION(SIZE(PUT,1), SIZE(PUT,2), SIZE(PUT,3)) :: ZMEAN, ZWORK
!$acc declare present(ZMEAN,ZWORK)
!
INTEGER :: K_SCHEME
INTEGER :: IKU
INTEGER :: IWORK
#ifdef _OPENACC
INTEGER :: IZFPOS1, IZFPOS2, IZFPOS3
INTEGER :: IZFNEG1, IZFNEG2, IZFNEG3
INTEGER :: IZBPOS1, IZBPOS2, IZBPOS3
INTEGER :: IZBNEG1, IZBNEG2, IZBNEG3
INTEGER :: IZOMP1,  IZOMP2,  IZOMP3
INTEGER :: IZOMN1,  IZOMN2,  IZOMN3
#endif
!
!
#ifdef _OPENACC
CALL INIT_ON_HOST_AND_DEVICE(ZMEAN,1e90,'ADVECUVW_WENO_K::ZMEAN')
CALL INIT_ON_HOST_AND_DEVICE(ZWORK,2e90,'ADVECUVW_WENO_K::ZWORK')
#endif
!
!------------------------- ADVECTION OF MOMENTUM ------------------------------
!
!
TZHALO2_UT => TPHALO2LIST                   ! 1rst add3dfield in model_n
TZHALO2_VT => TPHALO2LIST%NEXT              ! 2nd  add3dfield in model_n
TZHALO2_WT => TPHALO2LIST%NEXT%NEXT         ! 3rst add3dfield in model_n
!
IKU=SIZE(PUT,3)
ZMEAN = 0.0
ZWORK=0.0
!      -------------------------------------------------------
!
SELECT CASE(KWENO_ORDER)
!
CASE(1) ! WENO 1
#ifndef _OPENACC
!
!  U component
!
  PRUS = PRUS - DXM(UP_UX(PUT,MXF(PRUCT)))
!
  PRUS = PRUS - DYF(UP_MY(PUT,MXM(PRVCT)))
!
  PRUS = PRUS - DZF(1,IKU,1,UP_MZ(PUT,MXM(PRWCT)))
!
! V component
!
  PRVS = PRVS - DXF(UP_MX(PVT,MYM(PRUCT)))
!
  PRVS = PRVS - DYM(UP_VY(PVT,MYF(PRVCT)))
!
  PRVS = PRVS - DZF(1,IKU,1,UP_MZ(PVT,MYM(PRWCT)))
!
! W component
!
  PRWS = PRWS - DXF(UP_MX(PWT,MZM(1,IKU,1,PRUCT)))
!
  PRWS = PRWS - DYF(UP_MY(PWT,MZM(1,IKU,1,PRVCT)))
!
  PRWS = PRWS - DZM(1,IKU,1,UP_WZ(PWT,MZF(1,IKU,1,PRWCT)))
#else
!
!  U component
!
  !PRUS = PRUS - DXM(UP_UX(PUT,MXF(PRUCT)))
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL UP_UX_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!
  !PRUS = PRUS - DYF(UP_MY(PUT,MXM(PRVCT)))
  CALL MXM_DEVICE(PRVCT,ZWORK)
  CALL UP_MY_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!
  !PRUS = PRUS - DZF(1,IKU,1,UP_MZ(PUT,MXM(PRWCT)))
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL UP_MZ_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!
! V component
!
  !PRVS = PRVS - DXF(UP_MX(PVT,MYM(PRUCT)))
  CALL MYM_DEVICE(PRUCT,ZWORK)
  CALL UP_MX_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRVS = PRVS - ZWORK
!$acc end kernels
!
  !PRVS = PRVS - DYM(UP_VY(PVT,MYF(PRVCT)))
  CALL MYF_DEVICE(PRVCT,ZWORK)
  CALL UP_VY_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRVS = PRVS - ZWORK
!$acc end kernels
!
  !PRVS = PRVS - DZF(1,IKU,1,UP_MZ(PVT,MYM(PRWCT)))
  CALL MYM_DEVICE(PRWCT,ZWORK)
  CALL UP_MZ_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRVS = PRVS - ZWORK
!$acc end kernels
!
! W component
!
  !PRWS = PRWS - DXF(UP_MX(PWT,MZM(1,IKU,1,PRUCT)))
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL UP_MX_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  !PRWS = PRWS - DYF(UP_MY(PWT,MZM(1,IKU,1,PRVCT)))
  CALL MZM_DEVICE(PRVCT,ZWORK)
  CALL UP_MY_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  !PRWS = PRWS - DZM(1,IKU,1,UP_WZ(PWT,MZF(1,IKU,1,PRWCT)))
  CALL MZF_DEVICE(1,IKU,1,PRWCT,ZWORK)
  CALL UP_WZ_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DZM_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
#endif
!
!
CASE(3) ! WENO 3
#ifndef _OPENACC
!
! U component
!
  ZWORK = MXF(PRUCT)
  CALL ADVEC_WENO_K_2_UX(HLBCX, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2)
  PRUS = PRUS - DXM(ZMEAN)
  
!   
  IF (.NOT.L2D) THEN
    ZWORK = MXM(PRVCT)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2)
    PRUS = PRUS - DYF(ZMEAN)
  END IF
!
  PRUS = PRUS - DZF(1,IKU,1,WENO_K_2_MZ(PUT, MXM(PRWCT)))
!
! V component
!
  IF (.NOT.L2D) THEN
    ZWORK = MYM(PRUCT)
    CALL ADVEC_WENO_K_2_MX(HLBCX, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2)
    PRVS = PRVS - DXF(ZMEAN)
!   
    ZWORK = MYF(PRVCT)
    CALL ADVEC_WENO_K_2_VY(HLBCY, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2)
    PRVS = PRVS - DYM(ZMEAN)
!
    PRVS = PRVS - DZF(1,IKU,1,WENO_K_2_MZ(PVT, MYM(PRWCT)))
  END IF
!
! W component
!
  ZWORK = MZM(1,IKU,1,PRUCT)
  CALL ADVEC_WENO_K_2_MX(HLBCX, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2)
  PRWS = PRWS - DXF(ZMEAN)
!
  IF (.NOT.L2D) THEN
    ZWORK = MZM(1,IKU,1,PRVCT)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2)
    PRWS = PRWS - DYF(ZMEAN)
  END IF
!
  PRWS = PRWS - DZM(1,IKU,1,WENO_K_2_WZ(PWT,MZF(1,IKU,1,PRWCT)))
#else
  CALL MNH_GET_ZT3D(IZFPOS1,IZFPOS2,IZFNEG1,IZFNEG2,IZBPOS1,IZBPOS2,IZBNEG1,IZBNEG2,IZOMP1,IZOMP2,IZOMN1,IZOMN2)
!
! U component
!
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_2_UX(HLBCX, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2%WEST, TZHALO2_UT%HALO2%EAST,             &
                         ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                         ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                         ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!   
  IF (.NOT.L2D) THEN
    CALL MXM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2%NORTH, TZHALO2_UT%HALO2%SOUTH,           &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRUS = PRUS - ZWORK
!$acc end kernels
  END IF
!
!  PRUS = PRUS - DZF(1,IKU,1,WENO_K_2_MZ(PUT, MXM(PRWCT)))
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL WENO_K_2_MZ(PUT, ZWORK, ZMEAN,                                                                  &
                   ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                   ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                   ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
  CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!
! V component
!
  IF (.NOT.L2D) THEN
    CALL MYM_DEVICE(PRUCT,ZWORK)
    CALL ADVEC_WENO_K_2_MX(HLBCX, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2%WEST, TZHALO2_VT%HALO2%EAST,             &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
    CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
!   
    CALL MYF_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_VY(HLBCY, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2%NORTH, TZHALO2_VT%HALO2%SOUTH,           &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
    CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
!
!    PRVS = PRVS - DZF(1,IKU,1,WENO_K_2_MZ(PVT, MYM(PRWCT)))
    CALL MYM_DEVICE(PRWCT,ZWORK)
    CALL WENO_K_2_MZ(PVT, ZWORK, ZMEAN,                                                                  &
                     ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                     ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                     ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
    CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
  END IF
!
! W component
!
!  ZWORK = MZM(1,IKU,1,PRUCT)
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_2_MX(HLBCX, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2%WEST, TZHALO2_WT%HALO2%EAST,             &
                         ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                         ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                         ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  IF (.NOT.L2D) THEN
!    ZWORK = MZM(1,IKU,1,PRVCT)
    CALL MZM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2%NORTH, TZHALO2_WT%HALO2%SOUTH,           &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRWS = PRWS - ZWORK
!$acc end kernels
  END IF
!
!  PRWS = PRWS - DZM(1,IKU,1,WENO_K_2_WZ(PWT,MZF(1,IKU,1,PRWCT)))
  CALL MZF_DEVICE(1,IKU,1,PRWCT,ZWORK)
  CALL WENO_K_2_WZ(PWT, ZWORK, ZMEAN,                                                                  &
                   ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), &
                   ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), &
                   ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2) )
  CALL DZM_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  CALL MNH_REL_ZT3D(IZFPOS1,IZFPOS2,IZFNEG1,IZFNEG2,IZBPOS1,IZBPOS2,IZBNEG1,IZBNEG2,IZOMP1,IZOMP2,IZOMN1,IZOMN2)
#endif
!
!
CASE(5) ! WENO 5
#ifndef _OPENACC
!
! U component
!
  ZWORK = MXF(PRUCT)
  CALL ADVEC_WENO_K_3_UX(HLBCX, PUT, ZWORK, ZMEAN) 
  CALL GET_HALO(ZMEAN)! Update HALO 
  PRUS = PRUS - DXM(ZMEAN)
!
  IF (.NOT.L2D) THEN! 3D Case
   ZWORK = MXM(PRVCT)     
   CALL ADVEC_WENO_K_3_MY(HLBCY, PUT, ZWORK, ZMEAN)
   CALL GET_HALO(ZMEAN)! Update HALO 
   PRUS = PRUS - DYF(ZMEAN)
  END IF
!
  ZMEAN = WENO_K_3_MZ(PUT, MXM(PRWCT))
  CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  PRUS = PRUS - DZF(1,IKU,1,ZMEAN) 
!
! V component, only called in 3D case
!
  IF (.NOT.L2D) THEN
!
    ZWORK = MYM(PRUCT)
    CALL ADVEC_WENO_K_3_MX(HLBCX, PVT, ZWORK, ZMEAN) 
    CALL GET_HALO(ZMEAN)! Update HALO 
    PRVS = PRVS - DXF(ZMEAN)
!
    ZWORK = MYF(PRVCT)
    CALL ADVEC_WENO_K_3_VY(HLBCY, PVT, ZWORK, ZMEAN)
    CALL GET_HALO(ZMEAN)! Update HALO 
    PRVS = PRVS - DYM(ZMEAN)
!
    ZMEAN = WENO_K_3_MZ(PVT, MYM(PRWCT))
    CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
    PRVS = PRVS - DZF(1,IKU,1,ZMEAN) 
!
  END IF
!
! W component
!
  ZWORK = MZM(1,IKU,1,PRUCT)
  CALL ADVEC_WENO_K_3_MX(HLBCX, PWT, ZWORK, ZMEAN)
  CALL GET_HALO(ZMEAN)! Update HALO
  PRWS = PRWS - DXF(ZMEAN)
!
  IF (.NOT.L2D) THEN! 3D Case
    ZWORK = MZM(1,IKU,1,PRVCT)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PWT, ZWORK, ZMEAN)
    CALL GET_HALO(ZMEAN)! Update HALO
    PRWS = PRWS - DYF(ZMEAN)
  END IF
!
  ZMEAN = WENO_K_3_WZ(PWT,MZF(1,IKU,1,PRWCT))
  CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  PRWS = PRWS - DZM(1,IKU,1,ZMEAN)
#else
PRINT *,'OPENACC: advecuvw_weno_k::KWENO_ORDER=5 being implemented'
  CALL MNH_GET_ZT3D(IZFPOS1,IZFPOS2,IZFPOS3,IZFNEG1,IZFNEG2,IZFNEG3,IZBPOS1, &
                    IZBPOS2,IZBPOS3,IZBNEG1,IZBNEG2,IZBNEG3,IZOMP1,IZOMP2,IZOMP3,IZOMN1,IZOMN2,IZOMN3)
!
! U component
!
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_3_UX(HLBCX, PUT, ZWORK, ZMEAN, &
                         ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                         ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                         ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                         ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                         ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                         ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!   
  IF (.NOT.L2D) THEN
  CALL MXM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PUT, ZWORK, ZMEAN, &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                           ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                           ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                           ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
    CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRUS = PRUS - ZWORK
!$acc end kernels
  END IF
!
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL WENO_K_3_MZ(PUT,ZWORK,ZMEAN, &
                   ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                   ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                   ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                   ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                   ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                   ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
  CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRUS = PRUS - ZWORK
!$acc end kernels
!
! V component
!
  IF (.NOT.L2D) THEN
    CALL MYM_DEVICE(PRUCT,ZWORK)
    CALL ADVEC_WENO_K_3_MX(HLBCX, PVT, ZWORK, ZMEAN, &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                           ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                           ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                           ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
    CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
!   
    CALL MYF_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_VY(HLBCY, PVT, ZWORK, ZMEAN, &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                           ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                           ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                           ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
    CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
!
    CALL MYM_DEVICE(PRWCT,ZWORK)
    CALL WENO_K_3_MZ(PVT,ZWORK,ZMEAN, &
                     ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                     ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                     ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                     ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                     ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                     ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
    CALL DZF_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
    PRVS = PRVS - ZWORK
!$acc end kernels
  END IF
!
! W component
!
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_3_MX(HLBCX, PWT, ZWORK, ZMEAN, &
                         ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                         ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                         ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                         ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                         ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                         ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  IF (.NOT.L2D) THEN
  CALL MZM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PWT, ZWORK, ZMEAN, &
                           ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                           ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                           ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                           ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                           ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                           ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRWS = PRWS - ZWORK
!$acc end kernels
  END IF
!
  CALL MZF_DEVICE(1,IKU,1,PRWCT,ZWORK)
  CALL WENO_K_3_WZ(PWT,ZWORK,ZMEAN, &
                   ZT3D(:,:,:,IZFPOS1), ZT3D(:,:,:,IZFPOS2), ZT3D(:,:,:,IZFPOS3), &
                   ZT3D(:,:,:,IZFNEG1), ZT3D(:,:,:,IZFNEG2), ZT3D(:,:,:,IZFNEG3), &
                   ZT3D(:,:,:,IZBPOS1), ZT3D(:,:,:,IZBPOS2), ZT3D(:,:,:,IZBPOS3), &
                   ZT3D(:,:,:,IZBNEG1), ZT3D(:,:,:,IZBNEG2), ZT3D(:,:,:,IZBNEG3), &
                   ZT3D(:,:,:,IZOMP1),  ZT3D(:,:,:,IZOMP2),  ZT3D(:,:,:,IZOMP3),  &
                   ZT3D(:,:,:,IZOMN1),  ZT3D(:,:,:,IZOMN2),  ZT3D(:,:,:,IZOMN3) )
  CALL DZM_DEVICE(1,IKU,1,ZMEAN,ZWORK)
!$acc kernels
  PRWS = PRWS - ZWORK
!$acc end kernels
!
  CALL MNH_REL_ZT3D(IZFPOS1,IZFPOS2,IZFPOS3,IZFNEG1,IZFNEG2,IZFNEG3,IZBPOS1, &
                    IZBPOS2,IZBPOS3,IZBNEG1,IZBNEG2,IZBNEG3,IZOMP1,IZOMP2,IZOMP3,IZOMN1,IZOMN2,IZOMN3)
#endif
!
!
END SELECT
!             ---------------------------------
!$acc update self(PRUS,PRVS,PRWS)
!
END SUBROUTINE ADVECUVW_WENO_K
